// Copyright (C) 2017 Keelan Lightfoot
// Copyright (C) 2007-2008 Board of Regents of the University of Wisconsin
//             System (Univ. of Wisconsin-Madison, Trace R&D Center)
// Copyright (C) 2007-2008 Omnitor AB
// Copyright (C) 2007-2008 Voiceriver Inc
//
// This library is free software; you can redistribute it and/or modify it
// under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation; either version 2.1 of the License, or (at
// your option) any later version.

package gortty

import (
	//"fmt"
	"context"
	"math"
)

const (
	demodStateWaitStart = iota
	demodStateSample
	demodStateWaitStop

	autobaudStateWaitStart = iota
	autobaudStateWaitZeroX
	autobaudStateWaitStop
	autobaudStateDisabled
)

// Demodulator is fun
type Demodulator struct {
	input         chan int16
	output        chan byte
	sigOutput     chan [3]float64
	minSigTimeout int // minimum signal time out counter
	samplesPerBit int
	dsp           *dsp
	settings      *ModemSettings
	sigOut        bool
	samplerate    int
	div           int
	inhibitOutput bool
	autobaud      struct {
		enabled  bool    // audio baud detection enable/disable flag
		estimate float64 // autobaud estimate
	}
}

// NewDemodulator is fun
func NewDemodulator(ctx context.Context, s *ModemSettings, input chan int16, output chan byte) *Demodulator {

	d := new(Demodulator)
	d.input = input
	d.output = output

	d.settings = s

	centerFrequency := math.Abs(float64(d.settings.OneFreq)+float64(d.settings.ZeroFreq)) / 2 * 3

	d.div = int(math.Floor(float64(d.settings.SampleRate) / float64(centerFrequency)))
	if d.div == 0 {
		d.div = 1
	}

	d.samplerate = d.settings.SampleRate / d.div

	d.samplesPerBit = int(float64(d.samplerate) / d.settings.Baud)
	d.autobaud.enabled = false

	d.dsp = newDsp(d.settings.OneFreq, d.settings.ZeroFreq, d.samplerate)

	d.autobaud.estimate = d.settings.Baud
	go d.run(ctx)
	return d

}

func (d *Demodulator) Inhibit(i bool) {
	d.inhibitOutput = i
}

func (d *Demodulator) run(ctx context.Context) {

	var nextState int
	var sample int  // down counter for sampling instant
	var bit int     // current bit being received
	var baudot byte // received code
	var zeroX int
	var remainders map[int]int

	state := demodStateWaitStart // current demod state machine state
	var autobaudState int
	if d.autobaud.enabled {
		autobaudState = autobaudStateWaitStart
	} else {
		autobaudState = autobaudStateDisabled
	}
	count := 0
	for {
		select {
		case <-ctx.Done():
			return
		case sam := <-d.input:

			count++
			if count%d.div > 0 {

				continue
			}
			//	d.dsp.demod(d.l.update(float64(sam)))
			d.dsp.demod(float64(sam))

			if d.sigOut {
				m0, m1, m2 := d.dsp.level()
				d.sigOutput <- [3]float64{m0, m1, m2}

			} else {
				//fmt.Println(d.dsp.level())
				nextState = state

				switch state {
				case demodStateWaitStart:
					// The demodulator sits in this state when it
					// is waiting for a start bit.  When a
					// "0" start-bit is detected that is
					// above a minimum threshold it transitions
					// out of this state.

					if d.dsp.space() {
						sample = int(float64(d.samplerate) / d.settings.Baud * 1.5)
						baudot = 0
						bit = 0
						nextState = demodStateSample
						//	fmt.Println("demodStateSample")

					}

				case demodStateSample:

					sample--
					if sample == 0 {
						if d.dsp.mark() {
							baudot |= (1 << uint(bit))
						}
						bit++
						if bit == d.settings.DataBits {
							// complete baudot char has been received
							sample = int(float64(d.samplerate) / d.settings.Baud)
							nextState = demodStateWaitStop
							if !d.inhibitOutput {
								d.output <- baudot
							}
						} else {
							/* sample next bit 1 bit-period later */
							sample = int(float64(d.samplerate) / d.settings.Baud)
						}
					}

					/* check for d.sampleRateK signal drop out */

					if !d.dsp.carrier() {
						d.minSigTimeout++

						/* If signal drops out for approx one bit period
						   assume we have lost d.sampleRateK signal */
						if d.minSigTimeout > d.samplesPerBit {
							nextState = demodStateWaitStart
							//fmt.Println("demodStateWaitStart")

						}
					} else {
						d.minSigTimeout = 0
					}

				case demodStateWaitStop:

					/* wait until we are in the middle of stop bit */

					sample--
					if sample == 0 {
						nextState = demodStateWaitStart
						//fmt.Println("demodStateWaitStart")
					}

				}

				state = nextState

				// auto baud detection state machine
				// This deviates quite a bit from the original C implementation.
				// Starts a counter at falling edge that marks the
				// beginning of the start bit.  Measures time to
				// next rising edge.  This time is presumably from
				// 1 to 6 bit periods long, so the remainder from divinging
				// the time span by a numbe of bit periods for different baud
				// rates is accumulated. The baud rate that divides the most
				// evenly into all measured bit periods is the estimated baud rate.

				nextState = autobaudState

				switch autobaudState {
				case autobaudStateWaitStart:
					/* start bit kicks off this state machine */
					if state == demodStateSample {
						nextState = autobaudStateWaitZeroX
						zeroX = 0
						remainders = map[int]int{45: 0, 47: 0, 50: 0, 75: 0}

					}
				case autobaudStateWaitZeroX:

					zeroX++

					if d.dsp.edgeDetected() {

						estimate := 45 // pick one as a starting point
						minRemainder := remainders[estimate]

						for b, t := range remainders {
							q := t + (zeroX % (d.samplerate / b))
							remainders[b] = q
							if q < minRemainder {
								minRemainder = q
								estimate = b
							}
						}

						d.autobaud.estimate = float64(estimate)
						nextState = autobaudStateWaitStop
					}

					// reset autobaud state machine, e.g. if signal drops out before
					// we get a zero crossing

					if state != demodStateSample {
						nextState = autobaudStateWaitStop
					}

				case autobaudStateWaitStop:
					if state == demodStateWaitStop {
						nextState = autobaudStateWaitStart
					}

				case autobaudStateDisabled:

				}
				autobaudState = nextState
			}
		}
	}
}
